@@ -1,26 +1,23 @@
-"""Implements the xonsh executer"""
-from __future__ import print_function, unicode_literals
+# -*- coding: utf-8 -*-
+"""Implements the xonsh executer."""
 import re
-import os
+import sys
 import types
 import inspect
 import builtins
-from collections import Iterable, Sequence, Mapping
+from collections import Mapping
 
 from xonsh import ast
 from xonsh.parser import Parser
-from xonsh.tools import subproc_toks
+from xonsh.tools import subproc_toks, find_next_break
 from xonsh.built_ins import load_builtins, unload_builtins
 
 
 class Execer(object):
     """Executes xonsh code in a context."""
 
-    def __init__(self,
-                 filename='<xonsh-code>',
-                 debug_level=0,
-                 parser_args=None,
-                 unload=True):
+    def __init__(self, filename='<xonsh-code>', debug_level=0, parser_args=None,
+                 unload=True, config=None, login=True, xonsh_ctx=None):
         """Parameters
         ----------
         filename : str, optional
@@ -31,6 +28,10 @@ class Execer(object):
             Arguments to pass down to the parser.
         unload : bool, optional
             Whether or not to unload xonsh builtins upon deletion.
+        config : str, optional
+            Path to configuration file.
+        xonsh_ctx : dict or None, optional
+            Xonsh xontext to load as builtins.__xonsh_ctx__
         """
         parser_args = parser_args or {}
         self.parser = Parser(**parser_args)
@@ -38,13 +39,13 @@ class Execer(object):
         self.debug_level = debug_level
         self.unload = unload
         self.ctxtransformer = ast.CtxAwareTransformer(self.parser)
-        load_builtins(execer=self)
+        load_builtins(execer=self, config=config, login=login, ctx=xonsh_ctx)
 
     def __del__(self):
         if self.unload:
             unload_builtins()
 
-    def parse(self, input, ctx, mode='exec'):
+    def parse(self, input, ctx, mode='exec', wrap_subprocs=True):
         """Parses xonsh code in a context-aware fashion. For context-free
         parsing, please use the Parser class directly.
         """
@@ -67,7 +68,11 @@ class Execer(object):
         # tokens for all of the Python rules. The lazy way implemented here
         # is to parse a line a second time with a $() wrapper if it fails
         # the first time. This is a context-free phase.
-        tree = self._parse_ctx_free(input, mode=mode)
+        if wrap_subprocs:
+            tree, input = self._parse_ctx_free(input, mode=mode)
+        else:
+            return self.parser.parse(input, filename=self.filename, mode=mode,
+                                     debug_level=(self.debug_level > 1))
         if tree is None:
             return None
 
@@ -81,7 +86,7 @@ class Execer(object):
         return tree
 
     def compile(self, input, mode='exec', glbs=None, locs=None, stacklevel=2,
-                filename=None):
+                filename=None, wrap_subprocs=True):
         """Compiles xonsh code into a Python code object, which may then
         be execed or evaled.
         """
@@ -92,13 +97,14 @@ class Execer(object):
             glbs = frame.f_globals if glbs is None else glbs
             locs = frame.f_locals if locs is None else locs
         ctx = set(dir(builtins)) | set(glbs.keys()) | set(locs.keys())
-        tree = self.parse(input, ctx, mode=mode)
+        tree = self.parse(input, ctx, mode=mode, wrap_subprocs=wrap_subprocs)
         if tree is None:
             return None  # handles comment only input
         code = compile(tree, filename, mode)
         return code
 
-    def eval(self, input, glbs=None, locs=None, stacklevel=2):
+    def eval(self, input, glbs=None, locs=None, stacklevel=2,
+                wrap_subprocs=True):
         """Evaluates (and returns) xonsh code."""
         if isinstance(input, types.CodeType):
             code = input
@@ -107,12 +113,14 @@ class Execer(object):
                                 glbs=glbs,
                                 locs=locs,
                                 mode='eval',
-                                stacklevel=stacklevel)
+                                stacklevel=stacklevel,
+                                wrap_subprocs=wrap_subprocs)
         if code is None:
             return None  # handles comment only input
         return eval(code, glbs, locs)
 
-    def exec(self, input, mode='exec', glbs=None, locs=None, stacklevel=2):
+    def exec(self, input, mode='exec', glbs=None, locs=None, stacklevel=2,
+                wrap_subprocs=True):
         """Execute xonsh code."""
         if isinstance(input, types.CodeType):
             code = input
@@ -121,24 +129,12 @@ class Execer(object):
                                 glbs=glbs,
                                 locs=locs,
                                 mode=mode,
-                                stacklevel=stacklevel)
+                                stacklevel=stacklevel,
+                                wrap_subprocs=wrap_subprocs)
         if code is None:
             return None  # handles comment only input
         return exec(code, glbs, locs)
 
-    def _find_next_break(self, line, mincol):
-        if mincol >= 1:
-            line = line[mincol:]
-        if ';' not in line:
-            return None
-        maxcol = None
-        self.parser.lexer.input(line)
-        for tok in self.parser.lexer:
-            if tok.type == 'SEMI':
-                maxcol = tok.lexpos + mincol + 1
-                break
-        return maxcol
-
     def _parse_ctx_free(self, input, mode='exec'):
         last_error_line = last_error_col = -1
         parsed = False
@@ -148,7 +144,7 @@ class Execer(object):
                 tree = self.parser.parse(input,
                                          filename=self.filename,
                                          mode=mode,
-                                         debug_level=self.debug_level)
+                                         debug_level=(self.debug_level > 1))
                 parsed = True
             except IndentationError as e:
                 if original_error is None:
@@ -185,20 +181,37 @@ class Execer(object):
                     curr_indent = len(lines[idx]) - len(lines[idx].lstrip())
                     if prev_indent == curr_indent:
                         raise original_error
-                maxcol = self._find_next_break(line, last_error_col)
-                sbpline = subproc_toks(line,
-                                       returnline=True,
-                                       maxcol=maxcol,
-                                       lexer=self.parser.lexer)
+                lexer = self.parser.lexer
+                maxcol = find_next_break(line, mincol=last_error_col,
+                                         lexer=lexer)
+                sbpline = subproc_toks(line, returnline=True,
+                                       maxcol=maxcol, lexer=lexer)
                 if sbpline is None:
-                    # subprocess line had no valid tokens, likely because
-                    # it only contained a comment.
-                    del lines[idx]
-                    last_error_line = last_error_col = -1
-                    input = '\n'.join(lines)
-                    continue
+                    # subprocess line had no valid tokens,
+                    if len(line.partition('#')[0].strip()) == 0:
+                        # likely because it only contained a comment.
+                        del lines[idx]
+                        last_error_line = last_error_col = -1
+                        input = '\n'.join(lines)
+                        continue
+                    else:
+                        # or for some other syntax error
+                        raise original_error
+                elif sbpline[last_error_col:].startswith('![![') or \
+                     sbpline.lstrip().startswith('![!['):
+                    # if we have already wrapped this in subproc tokens
+                    # and it still doesn't work, adding more won't help
+                    # anything
+                    raise original_error
                 else:
+                    if self.debug_level:
+                        msg = ('{0}:{1}:{2}{3} - {4}\n'
+                               '{0}:{1}:{2}{3} + {5}')
+                        mstr = '' if maxcol is None else ':' + str(maxcol)
+                        msg = msg.format(self.filename, last_error_line,
+                                         last_error_col, mstr, line, sbpline)
+                        print(msg, file=sys.stderr)
                     lines[idx] = sbpline
                 last_error_col += 3
                 input = '\n'.join(lines)
-        return tree
+        return tree, input
