@@ -1,6 +1,7 @@
-"""
-Tools to open ``*.py`` files as Unicode, using the encoding specified within the
-file, as per PEP 263.
+# -*- coding: utf-8 -*-
+"""Tools to open ``*.py`` files as Unicode.
+
+Uses the encoding specified within the file, as per PEP 263.
 
 Much of the code is taken from the tokenize module in Python 3.2.
 
@@ -11,14 +12,10 @@ This file was forked from the IPython project:
 * Copyright (c) 2001, Janko Hauser <jhauser@zscout.de>
 * Copyright (c) 2001, Nathaniel Gray <n8gray@caltech.edu>
 """
-from __future__ import absolute_import
-
 import io
-import re
 import os.path
-from io import TextIOWrapper, BytesIO
+import re
 
-from xonsh.tools import unicode_type
 
 cookie_re = re.compile(r"coding[:=]\s*([-\w.]+)", re.UNICODE)
 cookie_comment_re = re.compile(r"^\s*#.*coding[:=]\s*([-\w.]+)", re.UNICODE)
@@ -84,12 +81,12 @@ except ImportError:
                 codec = lookup(encoding)
             except LookupError:
                 # This behaviour mimics the Python interpreter
-                raise SyntaxError("unknown encoding: " + encoding)
+                raise SyntaxError("unknown encoding: " + encoding, filename='<file>')
 
             if bom_found:
                 if codec.name != 'utf-8':
                     # This behaviour mimics the Python interpreter
-                    raise SyntaxError('encoding problem: utf-8')
+                    raise SyntaxError('encoding problem: utf-8', filename='<file>')
                 encoding += '-sig'
             return encoding
 
@@ -128,7 +125,7 @@ except ImportError:
         buf = io.open(filename, 'rb')  # Tweaked to use io.open for Python 2
         encoding, lines = detect_encoding(buf.readline)
         buf.seek(0)
-        text = TextIOWrapper(buf, encoding, line_buffering=True)
+        text = io.TextIOWrapper(buf, encoding, line_buffering=True)
         text.mode = 'r'
         return text
 
@@ -141,10 +138,10 @@ def source_to_unicode(txt, errors='replace', skip_encoding_cookie=True):
     txt can be either a bytes buffer or a string containing the source
     code.
     """
-    if isinstance(txt, unicode_type):
+    if isinstance(txt, str):
         return txt
     if isinstance(txt, bytes):
-        buf = BytesIO(txt)
+        buf = io.BytesIO(txt)
     else:
         buf = txt
     try:
@@ -152,7 +149,7 @@ def source_to_unicode(txt, errors='replace', skip_encoding_cookie=True):
     except SyntaxError:
         encoding = "ascii"
     buf.seek(0)
-    text = TextIOWrapper(buf, encoding, errors=errors, line_buffering=True)
+    text = io.TextIOWrapper(buf, encoding, errors=errors, line_buffering=True)
     text.mode = 'r'
     if skip_encoding_cookie:
         return u"".join(strip_encoding_cookie(text))
